#!/usr/bin/python

import datetime, gobject, ical, pytz, urllib2
from ubot.helper import UbotCommander, UbotResponder, main

def _bool(data):
    return data.lower() in ('1', 'true', 'yes', 'on')

def parse_tz(tz):
    if not tz:
        return pytz.UTC
    if 'gmt' in tz:
        return # Avoid confusion, gmt != utc
    # Exact matches
    for z in pytz.common_timezones:
        if tz.lower() == z.lower():
            return pytz.timezone(z)
    # Partial matches
    for z in pytz.common_timezones:
        if z.lower().endswith('/' + tz.lower().replace(' ','_')):
            return pytz.timezone(z)

class AgendaHelper(UbotCommander, UbotResponder):
    commands = {
        'now':      'now',
        'time':     'now',
        'schedule': 'schedule',
        'update':   'update',
    }

    def init(self, **kwargs):
        super(AgendaHelper, self).init(**kwargs)
        self.sources = {}
        gobject.timeout_add(60, self.update_sources)
        gobject.timeout_add(60, self.update_topics)
        self.update_sources()
        self.update_topics()

    def update_sources(self):
        done = []
        for s in self.conf.sections():
            if self.conf.has_option(s, 'source'):
                url = self.conf.get(s, 'source')
                if url in done:
                    continue
                done.append(url)
                try:
                    data = urllib2.urlopen(url).read()
                except:
                    self.log("Unable to retrieve %s" % url)
                    self.sources[url] = []
                    continue
                try:
                    data = ical.ICalReader(data)
                except:
                    self.log("Unable to parse %s" % url)
                    self.sources[url] = []
                    continue
                self.sources[url] = data.events

    def list_events(self, channel, amount, tz=pytz.UTC, notemplate=False):
        events = self.sources[self.conf.get(channel, 'source')]
        template = '%s'
        if  not notemplate and amount != 1:
            template = self.conf.get(channel, 'template')
        n = datetime.datetime.now(pytz.UTC)
        
        for e in events[:]:
            if e.seconds_ago() > 1800:
                events.remove(e)
            elif self.conf.get(channel, 'filter').lower() not in e.raw_data.lower():
                events.remove(e)
        if not len(events):
            return template % "No events scheduled"
        # The standard slack of 30 minutes after the meeting will be an
        # error if there are 2 conscutive meetings, so remove the first 
        # one in that case
        if len(events) > 1 and events[1].startDate < n:
            events.pop(0)
        preamble = ''
        if events[0].seconds_to_go() < 600:
            preamble = 'Current meeting: %s ' % events[0].summary.replace('Meeting','').strip()
            if amount == 1:
                return preamble
        if amount == 1:
            return 'Next meeting: %s in %s ' % (events[0].summary.replace('Meeting','').strip(), events[0].time_to_go())
        events = [x.schedule(tz) for x in events]
        return preamble + (template % ' | '.join(events))

    def update_topics(self):
        if not self.master:
            return
        for s in self.channels:
            if _bool(self.conf.get(s, 'topic')):
                current = self.channels[s].get_topic()
                new = self.list_events(s, 6)
                if current != new:
                    self.channels[s].set_topic(new)

    def update(self, prefix, target, params, msgparams):
        self.update_sources()
        self.update_topics()

    def now(self, prefix, target, params, tz):
        tz = parse_tz(tz)
        if not tz:
            self.reply("Unknown timezone", private=True, slow=True)
            return
        t = datetime.datetime.now(pytz.UTC)
        e = self.list_events(target, 1, tz)
        if self.sources[self.conf.get(target, 'source')] and \
           self.sources[self.conf.get(target, 'source')][0].is_on() and _bool(self.conf.get(target, 'silence')):
            self.reply("Please avoid bot commands during meetings", private=True)
            self.reply("Current time in %s: %s - %s" % (str(tz), t.astimezone(tz).strftime("%B %d %Y, %H:%M:%S"), e), private=True)
        else:
            self.reply("Current time in %s: %s - %s" % (str(tz), t.astimezone(tz).strftime("%B %d %Y, %H:%M:%S"), e))

    def schedule(self, prefix, target, params, tz):
        tz = parse_tz(tz)
        if not tz:
            self.reply("Unknown timezone" % tz, private=True, slow=True)
            return
        events = self.list_events(target, 6, tz, notemplate=True)
        if self.sources[self.conf.get(target, 'source')] and \
           self.sources[self.conf.get(target, 'source')][0].is_on() and _bool(self.conf.get(target, 'silence')):
            self.reply("Please avoid bot commands during meetings", private=True)
            self.reply("Schedule for %s: %s" % (str(tz), events), private=True)
        else:
            self.reply("Schedule for %s: %s" % (str(tz), events))

    def in_topic(self, prefix, target, params):
        if self.master and self.conf.get(target, 'topic'):
            self.reply("The topic of %s is managed by me, please don't change it manually" % target)

if __name__  == '__main__':
    main(AgendaHelper)
