import ubot.handlers
import ubot.irc
import ubot.exceptions
import ubot.rfc2812

import copy, datetime, dbus, gobject, os, re, sys, time
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
import cPickle as pickle
DBusGMainLoop(set_as_default=True)

class Ubot(dbus.service.Object):
    def __init__(self, config):
        self.config = config
        self.ircconnection = None
        self.busname = dbus.service.BusName("net.seveas.ubot." + config.busname, dbus.SessionBus())
        dbus.service.Object.__init__(self, dbus.SessionBus(), '/net/seveas/ubot/' + config.busname)
        self.logfd = open(self.config.logfile,'a')
        data = {}
        self.ignores = []
        if os.path.exists(self.config.datafile):
            fd = open(self.config.datafile)
            data = pickle.load(fd)
            fd.close()
        if os.path.exists(self.config.ignorefile):
            fd = open(self.config.ignorefile)
            self.ignores = [re.compile(x.strip()) for x in fd.readlines()]
            fd.close()
        self.saved_channels = data.get('saved_channels',set([]))
        self.tojoin = copy.deepcopy(self.saved_channels)
        self.saved_passwords = data.get('saved_passwords',{})
        self.am_master = False
        self.signal_master = False
        self.clear_data()
        self.last_recv = time.time()

    def clear_data(self):
        """Clear internal data, used when diconnected"""
        if hasattr(self, 'channels'):
            for c in self.channels.values():
                c.remove_from_connection()
                del(c)
        self.channels = {}
        self.prefixes = {}
        self.nick = None
        self.synced = len(self.saved_channels) == 0
        self.tojoin = copy.deepcopy(self.saved_channels)
        self.loggedin = False
        for p in self.config.peers:
            p.is_active = False
        self.set_master(False)

    def flush(self):
        obj = {
            'saved_channels': self.saved_channels,
            'saved_passwords': self.saved_passwords
        }
        fd = open(self.config.datafile, 'w')
        pickle.dump(obj,fd)
        fd.close()

    def set_master(self, value):
        if self.am_master == value:
            return
        if value:
            ubot.log.log("Assuming master role")
        else:
            ubot.log.log("Dropping master role")
        self.am_master = value
        self.signal_master = True
    
    def run(self):
        """Start the bot"""
        ubot.log.log("Connecting to %s" % str(self.config.servers))
        self.ircconnection = ubot.irc.IrcConnection(self.config.servers, self.handle_message)
        if not self.ircconnection.socket:
            ubot.log.log("Unable to connect")
            return
        gobject.timeout_add(60, self.clean_prefixes)
        gobject.timeout_add(60, self.ping)
        loop = gobject.MainLoop()
        loop.run()

    def clean_prefixes(self):
        """Clean old prefixes from the cache"""
        cutoff = time.time() - 900
        for p in self.prefixes.keys():
            if self.prefixes[p][1] + 900 < cutoff:
                self.prefixes.pop(p)
        return True

    def ping(self):
        """Ping the server and disconnect if it doesn't reply"""
        now = time.time()
        if now - self.last_recv > 90:
            self.ircconnection.reconnect()
            return
        elif now - self.last_recv > 60:
            self.ircconnection.send(ubot.irc.OutMessage('PING', ['lagometer']))

    def handle_message(self, message):
        """Handle an incoming/outgoing message"""
        # Log the message
        self.logfd.write("%s %s\n" % (datetime.datetime.now().strftime("%Y/%m/%d %H:%M"), repr(message)))
        if self.verbose:
            print "%s %s" % (datetime.datetime.now().strftime("%Y/%m/%d %H:%M"), repr(message))
            if self.verbose > 1 and message.raw:
                print ' ' * 17 + message.raw.strip()

        # Shortcut the short case, sent messages
        if isinstance(message, ubot.irc.OutMessage):
            self.message_sent(message.command, message.params)
            return

        self.lastrecv = time.time()
        # Should we ignore this message?
        if message.command in ('PRIVMSG', 'NOTICE') and not self.isme(message):
            for i in self.ignores:
                if i.match(message.prefix):
                    return

        # Send dbus signal
        self.message_received(message.prefix or '', message.command, message.target or '', message.params)

        # Handle the message internally
        if message.prefix and message.nick:
            self.prefixes[message.nick] = (message.prefix,time.time())
        if not self.loggedin:
            self.login()
        if 'handle_%s' % message.command.lower() in ubot.handlers.__dict__:
            ubot.handlers.__dict__['handle_%s' % message.command.lower()](message, self)
        elif message.ncommand and message.ncommand in ubot.irc.replies and \
             'handle_%s' % ubot.irc.replies[message.ncommand][4:].lower() in ubot.handlers.__dict__:
            ubot.handlers.__dict__['handle_%s' % ubot.irc.replies[message.ncommand][4:].lower()](message, self)

        # Are we synced yet
        if not self.synced and message.command in ('RPL_ENDOFNAMES', 'ERR_BANNEDFROMCHAN', 'ERR_INVITEONLYCHAN',
                                                   'ERR_CHANNELISFULL', 'ERR_NOSUCHCHANNEL', 'ERR_TOOMANYCHANNELS',
                                                   'ERR_BADCHANNELKEY', 'ERR_BADCHANMASK'):
            if message.params[0] in self.tojoin:
                self.tojoin.remove(message.params[0])
            if len(self.tojoin) == 0:
                self.synced = True
                ubot.log.log("Channel synchronization complete")
                self.handle_message(ubot.irc.InMessage("", "SYNCED", ["Sync complete"]))
        if self.signal_master:
            self.signal_master = False
            self.handle_message(ubot.irc.InMessage("", "MASTER", [str(int(self.am_master))]))

    @dbus.service.signal(dbus_interface='net.seveas.ubot', signature='sssas')
    def message_received(self, prefix, command, target, params):
        """Send the received message over the dbus"""
        pass

    @dbus.service.signal(dbus_interface='net.seveas.ubot', signature='sas')
    def message_sent(self, command, params):
        """Send the sent message over the dbus"""
        pass

    def login(self):
        """Identify to the server"""
        self.nick = self.config.nicks[0]
        ubot.log.log("Logging in as %s" % self.nick)
        if self.config.password:
            self.ircconnection.send(ubot.irc.OutMessage('PASS', [self.config.password]))
        self.ircconnection.send(ubot.irc.OutMessage('USER', [self.config.ident, '12', '*', self.config.realname]))
        self.ircconnection.send(ubot.irc.OutMessage('NICK', [self.config.nicks[0]]))
        self.loggedin = 1

    def isme(self, msg):
        """Determine if a message/prefix/nick is the bot's"""
        if isinstance(msg, ubot.irc.InMessage):
            return msg.nick == self.nick
        elif '!' in msg:
            return msg[:msg.index('!')] == self.nick
        return msg == self.nick

    # Here comes the DBUS api
    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='', out_signature='')
    def reload(self):
        """Reload data"""
        ubot.log.log("Reloading")
        #me = os.path.dirname(__file__) + os.pathsep
        #for m in sys.modules.values():
        #    if hasattr(m, __file__):
        #        if (os.path.dirname(m) + os.pathsep).startswith(me):
        #            reload(m)
        if os.path.exists(self.config.ignorefile):
            fd = open(self.config.ignorefile)
            self.ignores = [re.compile(x.strip()) for x in fd.readlines()]
            fd.close()

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='', out_signature='')
    def reopen_log(self):
        self.logfd.close()
        self.logfd = open(self.config.logfile,'a')

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='', out_signature='as')
    def get_channels(self):
        """Return a list of channels"""
        return self.channels.keys()

    # synced, master, nickname
    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='', out_signature='bbs')
    def get_info(self):
        return (self.synced, self.am_master, self.nick)

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='')
    def join(self, channel, password=''):
        """Join a channel, optionally with password"""
        if password:
            self.saved_passwords[channel] = password
            self.ircconnection.slowsend(ubot.irc.OutMessage('JOIN', [channel, password]))
        else:
            self.ircconnection.slowsend(ubot.irc.OutMessage('JOIN', [channel]))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='s', out_signature='')
    def quit(self, quitmessage):
        """Quit from the server"""
        self.ircconnection.fastsend(ubot.irc.OutMessage('QUIT', [quitmessage]))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='')
    def say(self, user, message):
        """Send a message to a user"""
        if len(user) < 1 or user[0] in '#@%':
            raise ubot.exceptions.InvalidTargetException('Can only send to people')
        self.ircconnection.send(ubot.irc.OutMessage('PRIVMSG', [user, message]))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='')
    def do(self, user, message):
        """Send a message to a user"""
        if len(user) < 1 or user[0] in '#@%':
            raise ubot.exceptions.InvalidTargetException('Can only send to people')
        self.ircconnection.send(ubot.irc.OutMessage('PRIVMSG', [user, '\x01ACTION ' + message + '\x01']))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='')
    def slowsay(self, user, message):
        """Send a message to a user"""
        if len(user) < 1 or user[0] in '#@%':
            raise ubot.exceptions.InvalidTargetException('Can only send to people')
        self.ircconnection.send(ubot.irc.OutMessage('PRIVMSG', [user, message]))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='')
    def slowdo(self, user, message):
        """Send a message to a user"""
        if len(user) < 1 or user[0] in '#@%':
            raise ubot.exceptions.InvalidTargetException('Can only send to people')
        self.ircconnection.send(ubot.irc.OutMessage('PRIVMSG', [user, '\x01ACTION ' + message + '\x01']))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='s', out_signature='')
    def nick(self, newnick):
        self.ircconnection.send(ubot.irc.OutMessage('NICK',[newnick]))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='ss', out_signature='', sender_keyword='sender')
    def log(self, ident, msg, sender=None):
        ubot.log.log(msg, ident + sender)
    
    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='sas', out_signature='')
    def rawmsg(self, cmd, args):
        """Send a raw message to the server"""
        self.ircconnection.send(ubot.irc.OutMessage(cmd, args))

    @dbus.service.method(dbus_interface='net.seveas.ubot',
                         in_signature='s', out_signature='as')
    def list_channels(self, nick):
        nick = ubot.rfc2812.IrcString(nick)
        return [x.name for x in self.channels.values() if nick in x.nicks]
