import ConfigParser, dbus, gobject, inspect, optparse, os, sys, thread
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

import ubot.util
from ubot.rfc2812 import IrcString

class MyConfigParser(ConfigParser.ConfigParser):
    def get(self, section, option, raw=False, vars=None):
        if not self.has_section(section):
            section = 'DEFAULT'
        return ConfigParser.ConfigParser.get(self, section, option, raw, vars)

class UbotHelper(object):
    def add_options(self, parser):
        parser.add_option('-n', '--name', dest='dbus_name', default='ubot', help='The busname of your bot')
        parser.add_option('-C', '--config', dest='config', default=None, help='Specify configfile', metavar='FILE')

    def init(self, **kwargs):
        self.conf = None
        if kwargs['config']:
            if os.path.exists(kwargs['config']):
                self.conf = MyConfigParser()
                self.conf.read(kwargs['config'])
        self.busname = kwargs['dbus_name']
        self.bot = dbus.SessionBus().get_object('net.seveas.ubot.' + self.busname,
                                                '/net/seveas/ubot/' + self.busname)

        self.channels = {}
        self.synced, self.master, self.nickname = self.bot.get_info()
        self.nickname = IrcString(self.nickname)
        if self.synced:
            self.in_synced(None,None,None)

        # Now that we're initialized, add signal receivers
        self.bot.connect_to_signal('message_sent', self.message_sent, dbus_interface='net.seveas.ubot')
        self.bot.connect_to_signal('message_received', self.message_received, dbus_interface='net.seveas.ubot')

    # Handle bus messages
    def message_sent(self, command, params):
        command = IrcString(command)
        params = [IrcString(x) for x in params]
        if hasattr(self, 'out_' + command.lower()):
            getattr(self, 'out_' + command.lower())(params)

    def message_received(self, prefix, command, target, params):
        prefix, command, target = IrcString(prefix), IrcString(command), IrcString(target)
        params = [IrcString(x) for x in params]
        command = command.replace('CMD_','').replace('RPL_','')
        if hasattr(self, 'in_' + command.lower()):
           getattr(self, 'in_' + command.lower())(prefix, target, params)

    def in_synced(self, prefix, target, params):
        # Request list of channels
        self.channels = dict([(x, dbus.SessionBus().get_object('net.seveas.ubot.' + self.busname, 
                               '/net/seveas/ubot/%s/channel/%s' % (self.busname, ubot.util.escape_object_path(x))))
                              for x in self.bot.get_channels()])

    def in_part(self, prefix, target, params):
        if prefix.startswith(self.nickname + '!'):
            x = self.channels.pop(target)
            del(x)

    def in_join(self, prefix, target, params):
        if prefix.startswith(self.nickname + '!'):
            self.channels[target] = dbus.SessionBus().get_object('net.seveas.ubot.' + self.busname, 
                                                        '/net/seveas/ubot/%s/channel/%s' % (self.busname, ubot.util.escape_object_path(target)))

    def in_nick(self, prefix, target, params):
        if prefix.startswith(self.nickname + '!'):
            self.nickname = params[0]

    def addressed(self, prefix, target, msg):
        return self.synced and self.master

    def log(self, msg):
        self.bot.log(self.name, msg)

class UbotResponder(UbotHelper):
    def add_options(self, parser):
        super(UbotResponder, self).add_options(parser)
        parser.add_option('-c', '--channels', dest='channels', default='', help='Channels this helper is active in', metavar='CHANNELS')

    def init(self, **kwargs):
        super(UbotResponder, self).init(**kwargs)
        self.active_channels = kwargs['channels'].split(',')
        self.respond_private = '' in self.channels or 'ALL' in self.channels

    def addressed(self, prefix, target, msg):
        if not super(UbotResponder, self).addressed(prefix, target, msg):
            return False
        if target == self.nickname:
            if not self.respond_private:
                return False
        else:
            if 'ALL' not in self.active_channels and target not in self.active_channels:
                return False
        return True

    def find_target(self):
        # Rewind through the stack to find the message_received funtion
        args = None
        try:
            frame = inspect.currentframe()
            while frame.f_code.co_name != 'message_received':
                _frame = frame.f_back
                del(frame) # Explicitely delete frame objects to avoid
                           # reference cycles as suggested by docs
                frame = _frame
            args = frame.f_locals
        except:
            return
        finally:
            del(frame)
        return args['prefix'], args['target']

    def reply(self, message, action=False, private=False, slow=False):
        prefix, target = self.find_target()
        if private:
            target = prefix[:prefix.index('!')]
        self.send(target, message, action, slow)

    def send(self, target, message, action=False, slow=False):
        if target.startswith('#'):
            if action:
                if slow:
                    self.channels[target].slowdo(message)
                else:
                    self.channels[target].do(message)
            else:
                if slow:
                    self.channels[target].slowsay(message)
                else:
                    self.channels[target].say(message)
        else:
            if action:
                if slow:
                    self.bot.slowdo(target, message)
                else:
                    self.bot.do(target, message)
            else:
                if slow:
                    self.bot.slowsay(target, message)
                else:
                    self.bot.say(target, message)

class UbotCommander(UbotHelper):
    def __init__(self):
        super(UbotCommander, self).__init__()
        self.__lastcommand = {}
        self.__lastargument = {}

    def add_options(self, parser):
        super(UbotCommander, self).add_options(parser)
        parser.add_option('-P', '--prefix',dest='prefix',metavar='PREFIX',help='Command prefix',default='@')

    def init(self, **kwargs):
        super(UbotCommander, self).init(**kwargs)
        self.prefix = kwargs['prefix']

    def addressed(self, prefix, target, msg):
        if not super(UbotCommander, self).addressed(prefix, target, msg):
            return False
        msg = msg.lstrip()
        for p in self.prefix:
            if msg.startswith(p):
                msg = msg[1:].lstrip()
                break
        else:
            return False

        # So, prefix was seen (FIXME: allow nickname as prefix)
        # Now for the commands
        if hasattr(self, 'commands'):
            _ = msg.split(None, 1)
            if len(_) == 2:
                command, arg = _
            else:
                command, arg = _[0], ''
            if command in self.commands:
                self.__lastcommand[thread.get_ident()], self.__lastargument[thread.get_ident()] = command, arg
                self.log("%s was called in %s by %s with argument %s" % (command, target, prefix, arg))
                return True
            else:
                return False

        # If a plugin doesn't define a command list, assume it'll do
        # its own argument handling
        return True

    def in_privmsg(self, prefix, target, params):
        if self.addressed(prefix, target, params[0]):
            command = getattr(self, self.commands[self.__lastcommand[thread.get_ident()]])
            command(prefix, target, params, self.__lastargument[thread.get_ident()])

def main(klass):
    c = klass()
    c.name = klass.__name__
    parser = optparse.OptionParser()
    parser.add_option('-H', '--host', dest='host', default='localhost', help='The host your ubot is running on')
    parser.add_option('-p', '--port', dest='port', default=11235, type=int, help='The port dbus-daemon listens on')
    c.add_options(parser)
    
    opts, args = parser.parse_args()
    os.environ['DBUS_SESSION_BUS_ADDRESS'] = 'tcp:host=%s,port=%d' % (opts.host, opts.port)
    c.init(**(opts.__dict__))
    c.log("helper joining the party")
    loop = gobject.MainLoop()
    loop.run()
