from django.db import models
from django.core import validators

def alias_or_value(field_data, all_data):
    if all_data['value'] and all_data['alias_of']:
        raise validators.ValidationError("Must be either alias or factoid, not both")

class Factoid(models.Model):
    name = models.CharField("Name", maxlength=25)
    value = models.TextField("Value", maxlength=1024, blank=True, validator_list=[alias_or_value])
    channel = models.CharField("Channel", blank=True, maxlength=30, help_text="Leave blank to make it default")
    added_by = models.CharField("Author", maxlength=50, editable=False)
    added_at = models.DateTimeField("Added at", auto_now=True)
    popularity = models.IntegerField("Popularity", default=0, editable=False)
    alias_of = models.ForeignKey('self', verbose_name = "Alias of", blank=True, null=True, validator_list=[alias_or_value])

    def __str__(self):
        if self.channel:
            return '%s (%s)' % (self.name, self.channel)
        return self.name
    
    def __repr__(self):
        return '<Factoid %s>' % self.name
    
    def __hash__(self):
        return hash(self._get_pk_val())

    def msg(self):
        self.error = False
        return self.value
    
    class Meta:
        unique_together = (('name','channel'),)

    class Admin:
        list_display = ('name', 'channel', 'added_by', 'value', 'alias_of')
        search_fields = ('name', 'value')
        ordering = ('-popularity', '+name')

debian = ('etch','lenny','sid','sarge')
ubuntu = ('dapper','gutsy','hardy','feisty')
class SourcePackage(models.Model):
    packagename = models.CharField("Name", maxlength=64, db_index=True)
    version = models.CharField("Version", maxlength=64)
    release = models.CharField("Release", maxlength=16, db_index=True)

    def url(self):
        if self.release in debian:
            return 'http://packages.qa.debian.org/%s/%s.html' % (self.packagename[0], self.packagename) 
        elif self.release in ubuntu:
            return 'https://launchpad.net/ubuntu/+source/' + self.packagename
        return ''

    def msg(self):
        return "%s is a sourcepackage in %s, version %s %s" % (self.packagename, self.release, self.version, self.url())

    def __str__(self):
        return "%s (%s/%s)" % (self.packagename, self.version, self.release)

    class Admin:
        search_fields = ('packagename','source')
        list_filter = ('release',)
        list_display = ('packagename','release','version')

class BinaryPackage(models.Model):
    packagename = models.CharField("Name", maxlength=64, db_index=True)
    version = models.CharField("Version", maxlength=64)
    release = models.CharField("Release", maxlength=16, db_index=True)
    shortdesc = models.CharField("Name", maxlength=128)
    source = models.CharField("Source", maxlength=64, blank=True, null=True)
    component = models.CharField("Component", maxlength=32)
    priority = models.CharField("Priority", maxlength=16)

    def url(self):
        if self.release in debian:
            s = self.source or self.packagename
            return 'http://packages.qa.debian.org/%s/%s.html' % (s[0], s) 
        elif self.release in ubuntu:
            s = self.source or self.packagename
            return 'https://launchpad.net/ubuntu/+source/' + s
        return ''

    def msg(self):
        source = self.source and (' (Source: %s)' % self.source) or ''
        return "%s%s: %s. In component %s, is %s. Version %s (%s) %s" % (self.packagename, source, self.shortdesc, self.component, 
                                                                        self.priority, self.version, self.release, self.url())

    def __str__(self):
        return "%s (%s/%s)" % (self.packagename, self.version, self.release)

    class Admin:
        search_fields = ('packagename','source')
        list_filter = ('release','component')
        list_display = ('packagename','source','release','version','component')
